@@ -2,10 +2,11 @@ package Mojolicious::Plugin::OpenAPI;
 use Mojo::Base 'Mojolicious::Plugin';
 
 use JSON::Validator::OpenAPI;
+use Mojo::JSON;
 use Mojo::Util 'deprecated';
 use constant DEBUG => $ENV{MOJO_OPENAPI_DEBUG} || 0;
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 
 sub EXCEPTION { +{errors => [{message => 'Internal server error.', path => '/'}], status => 500} }
 sub NOT_FOUND { +{errors => [{message => 'Not found.',             path => '/'}], status => 404} }
@@ -26,10 +27,12 @@ sub register {
     $app->helper('openapi.spec'          => \&_helper_spec);
     $app->helper('reply.openapi'         => \&_reply);
     $app->hook(before_render => \&_before_render);
+    $app->renderer->add_handler(openapi => \&_render);
     push @{$app->renderer->classes}, __PACKAGE__;
   }
 
   $self->{log_level} = $ENV{MOJO_OPENAPI_LOG_LEVEL} || $config->{log_level} || 'warn';
+  $self->{renderer} = $config->{renderer} || sub { Mojo::JSON::encode_json($_[1]) };
   $self->_validator->schema($api_spec->data)->coerce($config->{coerce} // 1);
   $self->_add_routes($app, $api_spec, $config);
 }
@@ -48,7 +51,7 @@ sub _add_routes {
   $base_path =~ s!/$!!;
 
   push @{$app->defaults->{'openapi.base_paths'}}, $base_path;
-  $route->to({'openapi.api_spec' => $api_spec, 'openapi.object' => $self});
+  $route->to({handler => 'openapi', 'openapi.api_spec' => $api_spec, 'openapi.object' => $self});
 
   my $spec_route = $route->get->to(cb => \&_reply_spec);
   if (my $spec_route_name = $config->{spec_route_name} || $api_spec->get('/x-mojo-name')) {
@@ -144,9 +147,10 @@ sub _log {
 }
 
 sub _reply {
-  my ($c, $status, $output) = @_;
-  my $self = $c->stash('openapi.object');
-  my $format = $c->stash('format') || 'json';
+  my $c      = shift;
+  my $status = ref $_[0] ? 200 : shift;
+  my $output = shift;
+  my @args   = @_;
 
   if (UNIVERSAL::isa($output, 'Mojo::Asset')) {
     my $h = $c->res->headers;
@@ -158,13 +162,28 @@ sub _reply {
     return $c->reply->asset($output);
   }
 
-  my @errors = $self->_validator->validate_response($c, $c->openapi->spec, $status, $output);
-  return $c->render($format => $output, status => $status) unless @errors;
-
-  $self->_log($c, '>>>', \@errors);
-  $c->render($format => {errors => \@errors, status => 500}, status => 500);
+  push @args, status => $status if $status;
+  return $c->render(@args, openapi => $output);
 }
 
+sub _render {
+  my ($renderer, $c, $output, $options) = @_;
+  my $res  = $c->stash('openapi')        or return;
+  my $self = $c->stash('openapi.object') or return;
+  my $status = $c->stash('status') || 200;
+  my $v = $self->_validator;
+
+  $c->stash->{format} ||= 'json';
+
+  if (my @errors = $v->validate_response($c, $c->openapi->spec, $status, $res)) {
+    $self->_log($c, '>>>', \@errors);
+    $c->stash(status => 500);
+    $$output = $self->{renderer}->($c, {errors => \@errors, status => 500});
+  }
+  else {
+    $$output = $self->{renderer}->($c, $res);
+  }
+}
 
 sub _reply_spec {
   my $c      = shift;
@@ -177,6 +196,7 @@ sub _reply_spec {
 
   return $c->render(json => $spec) unless $format eq 'html';
   return $c->render(
+    handler   => 'ep',
     template  => 'mojolicious/plugin/openapi/layout',
     esc       => sub { local $_ = shift; s/\W/-/g; $_ },
     serialize => \&_serialize,
@@ -239,7 +259,7 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
     my $data = {body => $c->validation->param("body")};
 
     # Validate the output response and render it to the user agent
-    $c->reply->openapi(200 => $data);
+    $c->render(openapi => $data);
   }, "echo";
 
   # Load specification and start web server
@@ -339,17 +359,9 @@ L</SYNOPSIS> for example usage.
 
 =head2 reply.openapi
 
-  $c->reply->openapi($status => $output);
+Will be DEPRECATED in favor of:
 
-Will L<validate|/openapi.validate> C<$output> before passing it on to
-L<Mojolicious::Controller/render>. Note that C<$output> will be passed on using
-the L<format|Mojolicious::Guides::Rendering/Content type> key in stash, which
-defaults to "json". This also goes for L<auto-rendering|/Controller>. Example:
-
-  my $format = $c->stash("format") || "json";
-  $c->render($format => \%output);
-
-C<$status> is a HTTP status code.
+  $c->render(openapi => $output);
 
 =head1 METHODS
 
@@ -385,6 +397,16 @@ C<log_level> is used when logging invalid request/response error messages.
 
 Default: "warn".
 
+=item * renderer
+
+Holds a code ref that can be used to render the response. Should return
+a plain string of data. The default is:
+
+  sub {
+    my ($c, $data) = @_;
+    return Mojo::JSON::encode_json($data);
+  }
+
 =item * route
 
 C<route> can be specified in case you want to have a protected API. Example:
